@@ -3,12 +3,10 @@ from typing import Callable, List, Sequence, Union
 from collections import defaultdict
 
 import sympy as sp
-from sympy.codegen.rewriting import optims_c99, optimize
-from sympy.codegen.rewriting import ReplaceOptim
 
 from pystencils.assignment import Assignment
-from pystencils.astnodes import Node, SympyAssignment
-from pystencils.field import AbstractField, Field
+from pystencils.astnodes import Node
+from pystencils.field import Field
 from pystencils.sympyextensions import subs_additive, is_constant, recursive_collect
 
 
@@ -164,7 +162,7 @@ def add_subexpressions_for_sums(ac):
     for eq in ac.all_assignments:
         search_addends(eq.rhs)
 
-    addends = [a for a in addends if not isinstance(a, sp.Symbol) or isinstance(a, AbstractField.AbstractAccess)]
+    addends = [a for a in addends if not isinstance(a, sp.Symbol) or isinstance(a, Field.Access)]
     new_symbol_gen = ac.subexpression_symbol_generator
     substitutions = {addend: new_symbol for new_symbol, addend in zip(new_symbol_gen, addends)}
     return ac.new_with_substitutions(substitutions, True, substitute_on_lhs=False)
@@ -226,23 +224,30 @@ def apply_on_all_subexpressions(operation: Callable[[sp.Expr], sp.Expr]):
     f.__name__ = operation.__name__
     return f
 
+# TODO Markus
+# make this really work for Assignmentcollections
+# this function should ONLY evaluate
+# do the optims_c99 elsewhere optionally
 
-def apply_sympy_optimisations(assignments):
-    """ Evaluates constant expressions (e.g. :math:`\\sqrt{3}` will be replaced by its floating point representation)
-        and applies the default sympy optimisations. See sympy.codegen.rewriting
-    """
-
-    # Evaluates all constant terms
-    evaluate_constant_terms = ReplaceOptim(lambda e: hasattr(e, 'is_constant') and e.is_constant and not e.is_integer,
-                                           lambda p: p.evalf())
-
-    sympy_optimisations = [evaluate_constant_terms] + list(optims_c99)
-
-    assignments = [Assignment(a.lhs, optimize(a.rhs, sympy_optimisations))
-                   if hasattr(a, 'lhs')
-                   else a for a in assignments]
-    assignments_nodes = [a.atoms(SympyAssignment) for a in assignments]
-    for a in chain.from_iterable(assignments_nodes):
-        a.optimize(sympy_optimisations)
-
-    return assignments
+# def apply_sympy_optimisations(ac: AssignmentCollection):
+#     """ Evaluates constant expressions (e.g. :math:`\\sqrt{3}` will be replaced by its floating point representation)
+#         and applies the default sympy optimisations. See sympy.codegen.rewriting
+#     """
+#
+#     # Evaluates all constant terms
+#
+#     assignments = ac.all_assignments
+#
+#     evaluate_constant_terms = ReplaceOptim(lambda e: hasattr(e, 'is_constant') and e.is_constant and not e.is_integer,
+#                                            lambda p: p.evalf())
+#
+#     sympy_optimisations = [evaluate_constant_terms] + list(optims_c99)
+#
+#     assignments = [Assignment(a.lhs, optimize(a.rhs, sympy_optimisations))
+#                    if hasattr(a, 'lhs')
+#                    else a for a in assignments]
+#     assignments_nodes = [a.atoms(SympyAssignment) for a in assignments]
+#     for a in chain.from_iterable(assignments_nodes):
+#         a.optimize(sympy_optimisations)
+#
+#     return AssignmentCollection(assignments)
