@@ -9,6 +9,7 @@
 """
 import pystencils
 import pystencils.astnodes
+import pystencils.config
 
 
 def test_source_code_comment():
@@ -19,7 +20,7 @@ def test_source_code_comment():
         {a.center(): b[0, 2] + b[0, 0]}, {}
     )
 
-    config = pystencils.CreateKernelConfig(target=pystencils.Target.CPU)
+    config = pystencils.config.CreateKernelConfig(target=pystencils.Target.CPU)
     ast = pystencils.create_kernel(assignments, config=config)
 
     ast.body.append(pystencils.astnodes.SourceCodeComment("Hallo"))
